/******************************************************************************
 * program:     rasimg library                                                *
 * function:    Executable for image comparison.                              *
 * modul:       gcomp.cc                                                      *
 * licency:     GPL or LGPL                                                   *
 ******************************************************************************/
#include <stddef.h>
#include <stdio.h>

#include <stdlib.h>
//#include <ctype.h>
#include <string.h>

#include "typedfs.h"
#include "raster.h"
#include "imageop.h"

#include <matrix.h>


static char Verbosity = 0;
double Epsilon = 1e-6;


const char *GetExtension(const char *FullName)
{
const char *NameStep;
int LastPos;

if(FullName==NULL) return(NULL);

LastPos=strlen(FullName);
NameStep=FullName+LastPos+1;
while(NameStep>FullName)
	{
	NameStep--;

	if(*NameStep=='/' || *NameStep=='\\' || *NameStep==':')
		break;
	if(*NameStep=='.')
		return(NameStep);
	}

return(FullName+LastPos);
}


/////////////////////Main CODE starts here/////////////////////////////////


/** Compare two Images. */
long Compare(const Image *pImg1, const Image *pImg2)
{
WORD y,x;
long Difference;
Raster2DAbstract *Ras1;
Raster2DAbstract *Ras2;
Raster1DAbstract *Ras_1;
Raster1DAbstract *Ras_2;

 Difference=0;

 Ras1 = pImg1->Raster;
 Ras2 = pImg2->Raster;

 if(Ras1==NULL || Ras2==NULL) return -1;
 if(Ras1->Data2D==NULL || Ras2->Data2D==NULL) return -2;
 if(Ras1->GetSize1D()!=Ras2->GetSize1D() || Ras1->Size2D!=Ras2->Size2D)
   {
   if(Verbosity>=0)
     {
     printf("  Wrong image size %dx%d compared to %dx%d ",
		Ras1->GetSize1D(),Ras1->Size2D, Ras2->GetSize1D(),Ras2->Size2D);
     }
   return -3;
   }

/* if AlineProc<>nil then
	AlineProc^.InitPassing(Obr1.y,'Comparing 2 images'); */

 if(Verbosity>=0)
   printf("  Image types I1:%X (pl%d%s); I2:%X (pl%d%s) ", pImg1->ImageType(),Ras1->GetPlanes(), (pImg1->Palette==NULL)?"":"IDX",
			pImg2->ImageType(),Ras2->GetPlanes(), (pImg2->Palette==NULL)?"":"IDX");

 if(pImg1->ImageType()==ImageTrueColor || pImg2->ImageType()==ImageTrueColor ||
    (pImg1->ImageType()==ImagePalette || pImg2->ImageType()==ImagePalette) )
   {		// True color process is required
   for(y=0; y<Ras1->Size2D; y++)
    {
    for(x=0; x<Ras1->GetSize1D(); x++)
      {
      if(pImg1->GetPixelRGB(x,y) != pImg2->GetPixelRGB(x,y))
	{
	//DWORD d = pImg1->GetPixelRGB(x,y);
        //DWORD e = pImg2->GetPixelRGB(x,y);

	if(Difference<10)
	  printf("\n [%d %d]: %lXh %lXh ",x,y,pImg1->GetPixelRGB(x,y),pImg2->GetPixelRGB(x,y));
	if(Difference==10)
          printf("\nAnd more ...");
        Difference++;
        }
      }
    }
   }
 else if(Ras1->GetPlanes()<=32 && Ras1->GetPlanes()>=-8 && Ras2->GetPlanes()<=32 && Ras2->GetPlanes()>=-8)
   {
   for(y=0; y<Ras1->Size2D; y++)
    {
    Ras_1 = Ras1->GetRowRaster(y);
    Ras_2 = Ras2->GetRowRaster(y);
    if(Ras_1==NULL || Ras_2==NULL) 
      {
      printf("\nEmpty row for line %d",y);
      Difference++;
      continue;
      }

    for(x=0; x<Ras1->GetSize1D(); x++)
      {
      if(Ras_1->GetValue1D(x) != Ras_2->GetValue1D(x))
        {
	if(Difference<10)
          printf("\n [%d %d]: %X %X ",x,y,pImg1->GetPixel(x,y),pImg2->GetPixel(x,y));
	if(Difference==10)
          printf("\nAnd more ...");
	Difference++;
        }
      }
    }
  }
 else
   {
   double MaxDelta = 0;
   for(y=0; y<Ras1->Size2D; y++)
    for(x=0; x<Ras1->GetSize1D(); x++)
      {
      double R1 = Ras1->GetValue2Dd(x,y);
      double R2 = Ras2->GetValue2Dd(x,y);
      double delta = fabs(R1-R2);
      if(delta > MaxDelta) MaxDelta=delta;
      if(delta > Epsilon)
        {
	if(Difference<10)
	  printf("\n [%d %d]: %f %f; eps=%f ", x, y, R1, R2, R1-R2);
	if(Difference==10)
          printf("\nAnd more ...");
	Difference++;
        }
      }
    if(Verbosity>=0) printf("\nMaximal delta = %g", MaxDelta);
  }

return Difference;
}


long Compare(const Image & Img1, const Image & Img2)
{
const Image *pI1, *pI2;
long diffs = 0;
int scene=0;

  pI1 = &Img1;
  pI2 = &Img2;
  do
  {
    if(Verbosity>=0)
      if(Img1.Next!=NULL || Img2.Next!=NULL)
        printf("\nScene #%u ", scene++);
    diffs += Compare(pI1,pI2);
    pI1 = pI1->Next;
    pI2 = pI2->Next;
  } while(pI1!=NULL && pI2!=NULL);

  if((pI1!=NULL || pI2!=NULL) && diffs==0)
  {
    printf("\nUnalligned amount of scenes");
    diffs = -1000;
  }
return diffs;
}


void PrintHeader(void)
{
char HdrPrinted=0;
  if(Verbosity>=0 && HdrPrinted==0)
    {
    printf("\n\r <<< GComp >>> Compare two graphical images (c)1997-2022 F&TSoft, version %u.%u\n", (RAS_IMG_VERSION>>8), (RAS_IMG_VERSION & 0xFF));
    HdrPrinted=1;
    }
}


int main(int argc,char *argv[])
{
Image I1, I2;
const char *Img1=NULL, *Img2=NULL;
const char *s;
TImageFileHandler *LoadPicture_XXX;
int i;
bool Conv2Gray = false;
bool ChannelIDX = false;
int Separate = -1;
int Error = 0;

/* Parse Arguments */
for(i=1; i<argc; i++)
  {
  s=argv[i];
  if(s==NULL) continue;
  if( *s==0 ) continue;

  if(*s=='-' || *s=='/')
    {
    if(strcmp(s+1,"help")==0 || strcmp(s+1,"-help")==0 || strcmp(s+1,"?")==0)
      {
        PrintHeader();
        printf("Supported file formats are:\n");
        for(LoadPicture_XXX=TImageFileHandler::first();LoadPicture_XXX!=NULL;LoadPicture_XXX=LoadPicture_XXX->next())
	  {
	  printf("%s %s%s; ",LoadPicture_XXX->extension(),
			     LoadPicture_XXX->LoadPicture==NULL?"":"R",
			     LoadPicture_XXX->SavePicture==NULL?"":"W");
	  }
	printf("\n/s            - silent mode"
	       "\n/v            - verbose"
		//"\n/NoScale      - means truncating image data"
	       "\n/Epsilon  val - Tolerated delta for operations with doubles"
	       "\n/Gray         - discards all color informations"
	       "\n/IDX          - use index channel"
	       "\n/Separate[R|G|B] - remove one channel from true color images"
	       "\n");
	return(0);
      }

     if(strcmp(s+1,"Epsilon") == 0)
     {
       ++i;
       if(i>=argc) break;
       Epsilon = atof(argv[i]);
       if(Epsilon <= 0)
       {
         printf("\n\rInvalid delta for double values: %f.", Epsilon);
         goto End;
       }
       continue;
     }
    if(strcmp(s+1,"s") == 0)
      {
      Verbosity = -1;
      continue;
      }
    if(strcmp(s+1,"v") == 0)
      {
      Verbosity = +1;
      continue;
      }
    if(strcmp(s+1,"Gray") == 0)
      {
      Conv2Gray = true;
      continue;
      }
    if(strcmp(s+1,"IDX") == 0)
      {
      ChannelIDX = true;
      continue;
      }
     if(strcmp(s+1,"SeparateR") == 0)
     {
       Separate = 0;
       continue;
     }
     if(strcmp(s+1,"SeparateG") == 0)
     {
       Separate = 1;
       continue;
     }
     if(strcmp(s+1,"SeparateB") == 0)
     {
       Separate = 2;
       continue;
     }
     if(strcmp(s+1,"Separate") == 0)
     {
       ++i;
       if(i>=argc) break;
       if(strcmp(s,"R")==0) Separate=0;
       else if(strcmp(s,"G")==0) Separate=1;
       else if(strcmp(s,"B")==0) Separate=2;
       else
       {
         printf("\n\rInvalid channel for separiation: %s.", s);
         goto End;
       }
       continue;
     }

    }

  if(Img1==NULL)
     Img1=s;
  else
     if(Img2==NULL) Img2=s;
  }
/* End of reading arguments */

  PrintHeader();
  if(Verbosity>=0) printf("\rLoading");
  //MarkTime=GetInterval();

  I1 = LoadPicture(Img1);
  if(I1.Raster==NULL && I1.VecImage==NULL)
	{
	printf("\n\rCannot load 1'st picture:\"%s\" Error %d",Img1,Error);
	Error = -11;
	goto End;
	}

  I2 = LoadPicture(Img2);
  if(I2.Raster==NULL && I2.VecImage==NULL)
	{
	printf("\n\rCannot load 2'nd picture:\"%s\" Error %d",Img2,Error);
	Error = -12;
	goto End;
	}

  if(Conv2Gray)
  {
    if(Verbosity>=0) printf("\rConverting to gray ");
    Convert2Gray(I1);
    Convert2Gray(I2);
  }

  if(ChannelIDX)
  {
    I1.AttachPalette(NULL);
    I2.AttachPalette(NULL);
  }

  if(Separate>=0)
  {
    if(Verbosity>=0) printf("\rSeparating channel %d ",Separate);
    I1 = Peel8bit(I1, Separate);
    I1.AttachPalette(NULL);
    I2 = Peel8bit(I2, Separate);
    I2.AttachPalette(NULL);
  }

  if(Verbosity>=0) printf("\rComparing ");
  i = Compare(I1,I2);
  if(i==0)
    {
    if(Verbosity>=0)
      printf("\nBoth images have same contents.\n\r");
    }
  else
    {
    Error = 1;
    if(i<0) printf("\nError %d occured during compare!\n\r",i);
       else printf("\nFound %d mismatches!\n\r",i);
    }
  //MarkTime=GetInterval();

End:
  return(Error);
}

